# Copyright 2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from attrs import define, field
import re
from slugify import slugify
from stop_words import get_stop_words


RE_ENGLISH_ORDINAL = re.compile(
    r"[0-9]+(st|nd|rd|th)|[a-z]+(ieth|enth)|first|second|third|fourth|fifth|sixth|eighth|ninth|twelfth"
)
STOPWORDS_VOLUMES = {
    "proceedings",
    "volume",
    "conference",
    "workshop",
    "annual",
    "meeting",
    "computational",
}


@define
class _StopWords:
    """This class exposes a (slugified) stop word list used for generating citation keys."""

    stopwords: set[str] = field(init=False, factory=set)
    loaded: bool = field(init=False, default=False)

    def load(self) -> None:
        self.stopwords = set(
            t for w in get_stop_words("en") for t in slugify(w).split("-")
        )
        self.loaded = True

    def contains(self, word: str) -> bool:
        """Check if a given word is contained in the stop word list.

        Parameters:
            word: The word to look up.

        Returns:
            True if `word` is a known stop word, False otherwise.
        """
        if not self.loaded:
            self.load()
        return word in self.stopwords

    def is_stopword(self, word: str) -> bool:
        """Alias for `self.contains`."""
        return self.contains(word)

    def is_volume_stopword(self, word: str) -> bool:
        """Check if a given word is considered a stop word in the context of volume titles.

        This will also return True on words like "proceedings" or "workshop", as well as ordinal expressions.

        Parameters:
            word: The word to look up.

        Returns:
            True if `word` is considered a volume-level stop word, False otherwise.
        """
        return (
            self.contains(word)
            or word in STOPWORDS_VOLUMES
            or RE_ENGLISH_ORDINAL.fullmatch(word) is not None
        )


StopWords = _StopWords()



# Copyright 2024 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for generating citation references."""

from __future__ import annotations

import citeproc
from citeproc import (
    Citation,
    CitationItem,
    CitationStylesBibliography,
    CitationStylesStyle,
)
from citeproc.source.json import CiteProcJSON
from pathlib import Path
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from ..collections import Paper
    from ..people import NameSpecification


class CitationStyleDict(dict[str | Path, Any]):
    """Dictionary mapping names of citation styles to their `citeproc.CitationStylesStyle` objects, auto-loading styles on first access."""

    def __getitem__(self, key: str | Path) -> CitationStylesStyle:
        style = dict.get(self, key)
        if style is None:
            if Path(key).is_file():
                # Assume that key is a filename
                filename = key
            else:
                # Assume that key is the name of a style in citeproc-py-styles
                raise ValueError(
                    "Names of citation styles are currently not supported; give the name to a CSL file instead."
                )
            if not Path(filename).is_file():
                raise KeyError(
                    f"Could not resolve '{key}' to a filename of a citation style"
                )
            style = CitationStylesStyle(filename)
            self.__setitem__(key, style)
        return style


citation_styles = CitationStyleDict()
"""Global object for accessing `citeproc.CitationStylesStyle` objects."""


def citeproc_render_html(
    citeproc_dict: dict[str, Any],
    style: str | Path,
    link_title: bool = True,
) -> str:
    """Render a bibliography entry with a given CSL style.

    Arguments:
        citeproc_dict: A dictionary with publication metadata as expected by CiteProcJSON.
        style: A path to a CSL file.
        link_title: If True, wraps the title in a link to the entry's URL.

    Returns:
        The bibliography entry as a single string with HTML markup.

    Note:
        The reason for returning a string is that this is what we get from citeproc-py's `render_bibliography()` function.  If the result was parsed with LXML, we could turn it into a proper [MarkupText][acl_anthology.text.MarkupText] object.  However, since the most common use case of this function requires the HTML-ified string, we do not do this here as it would introduce unnecessary overhead in this case.
    """
    source = CiteProcJSON([citeproc_dict])
    item = CitationItem(citeproc_dict["id"])
    bib = CitationStylesBibliography(
        citation_styles[style], source, citeproc.formatter.html
    )
    bib.register(Citation([item]))
    rendered_list = bib.style.render_bibliography([item])[0]
    if link_title:
        link_text = f'<a href="{citeproc_dict["URL"]}">{citeproc_dict["title"]}</a>'
        rendered_list = [
            str(x) if x != citeproc_dict["title"] else link_text for x in rendered_list
        ]
    return "".join(rendered_list)


def _format_names(names: list[NameSpecification]) -> str:
    match len(names):
        case 0:
            return "N.N."
        case 1:
            return names[0].name.as_first_last()
        case 2:
            return " and ".join(ns.name.as_first_last() for ns in names)
        case _:
            return f"{', '.join(ns.name.as_first_last() for ns in names[:-1])}, and {names[-1].name.as_first_last()}"


def _format_pages(pages: str) -> str:
    return pages.replace("--", "–").replace("-", "–")


def render_acl_citation(paper: Paper) -> str:
    """Render a bibliography entry in ACL style.

    Arguments:
        paper: The paper for which to generate the bibliography entry.

    Returns:
        The bibliography entry as a single string with HTML markup.

    Note:
        This function re-implements (parts of) the ACL citation style in pure Python, making it a much faster alternative to [citeproc_render_html][acl_anthology.utils.citation.citeproc_render_html].
    """
    if paper.authors:
        authors = _format_names(paper.authors)
    else:
        editors = paper.get_editors()
        if not editors:
            # No authors, no editors
            authors = ""
        else:
            authors = _format_names(editors)
            if not paper.is_frontmatter:
                authors = f"{authors} ({'eds.' if len(editors) > 1 else 'ed.'})"
    title = f'<a href="{paper.web_url}">{paper.title.as_text()}</a>'
    if paper.is_frontmatter:
        title = f"<i>{title}</i>"
    parent = []
    if paper.bibtype == "inproceedings":
        parent = [f"In <i>{paper.parent.title.as_text()}</i>"]
        if paper.pages:
            pages = _format_pages(paper.pages)
            parent.append(f", {'pages' if '–' in pages else 'page'} {pages}")
        if paper.address:
            parent.append(f", {paper.address}")
        if paper.publisher:
            parent.append(f". {paper.publisher}")
    elif paper.bibtype == "article":
        parent = [f"<i>{paper.get_journal_title()}</i>"]
        if paper.parent.journal_volume:
            parent.append(f", {paper.parent.journal_volume}")
            if (journal_issue := paper.get_issue()) is not None:
                parent.append(f"({journal_issue})")
            if paper.pages:
                parent.append(f":{_format_pages(paper.pages)}")
        elif paper.pages:
            parent.append(f", {_format_pages(paper.pages)}")
    else:
        if paper.publisher:
            parent.append(paper.publisher)
            if paper.address:
                parent.append(f", {paper.address}")

    if authors:
        citation = f"{authors}. {paper.year}. {title}."
    else:
        citation = f"{title}. {paper.year}."

    if parent:
        citation = f"{citation} {''.join(parent)}."

    return citation



# Copyright 2023-2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from attrs import define, field, validators as v, asdict
from pathlib import Path
from typing import Iterator, Optional, TYPE_CHECKING
import yaml

try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:  # pragma: no cover
    from yaml import Loader, Dumper  # type: ignore

from .utils.attrs import auto_validate_types
from .utils.ids import AnthologyIDTuple, build_id_from_tuple
from .containers import SlottedDict

if TYPE_CHECKING:
    from _typeshed import StrPath
    from .anthology import Anthology
    from .collections import Volume


@define(field_transformer=auto_validate_types)
class Venue:
    """A publication venue.

    Attributes:
        id: The venue ID, e.g. "acl".
        parent: The parent Anthology instance to which this venue belongs.
        acronym: The venue's acronym, e.g. "ACL".
        name: The venue's name.  Should _not_ contain any indications of specific events; i.e., "Workshop on...", _not_ "The 1st Workshop on..."
        path: The path of the YAML file representing this venue.
        is_acl: True if this is a venue organized or sponsored by the ACL.
        is_toplevel: True if this venue appears on the ACL Anthology's front page.
        item_ids: A list of volume IDs associated with this venue.
        oldstyle_letter: First letter of old-style Anthology IDs that is associated with this venue (e.g., "P" for ACL proceedings).
        url: A website URL for the venue.
    """

    id: str = field(converter=str)
    parent: Anthology = field(repr=False, eq=False)
    acronym: str = field(converter=str)
    name: str = field(converter=str)
    path: Path = field(converter=Path, eq=False)
    is_acl: bool = field(default=False, converter=bool)
    is_toplevel: bool = field(default=False, converter=bool)
    item_ids: list[AnthologyIDTuple] = field(
        factory=list,
        repr=lambda x: f"<list of {len(x)} AnthologyIDTuple objects>",
        eq=False,
    )
    oldstyle_letter: Optional[str] = field(
        default=None, validator=v.optional(v.matches_re("^[A-Z]$"))
    )
    url: Optional[str] = field(default=None, validator=v.optional(v.instance_of(str)))
    # TODO: Should we reconsider 'type'? Currently used to designate journals
    # at the venue level; but journals are also marked on the individual
    # volumes.
    type: Optional[str] = field(default=None, validator=v.optional(v.instance_of(str)))

    @classmethod
    def load_from_yaml(cls, path: StrPath, parent: Anthology) -> Venue:
        """Instantiates a venue from its YAML file.

        Arguments:
            path: The YAML file defining this venue.
            parent: The parent Anthology instance.

        Warning:
            Currently assumes that files are named `{venue_id}.yaml`.
        """
        path = Path(path)
        venue_id = path.name[:-5]
        with open(path, "r", encoding="utf-8") as f:
            kwargs = yaml.load(f, Loader=Loader)
        return cls(venue_id, parent=parent, path=path, **kwargs)

    def save(self, path: Optional[StrPath] = None) -> None:
        """Saves this venue as a YAML file.

        Arguments:
            path: The filename to save to. If None, defaults to `self.path`.
        """
        if path is None:
            path = self.path
        # Serialize everything except "id", "item_ids", "path", "parent" and default values
        values = asdict(
            self,
            filter=lambda a, v: a.name not in ("id", "item_ids", "parent", "path")
            and v != a.default,
        )
        with open(path, "w", encoding="utf-8") as f:
            yaml.dump(values, f, Dumper=Dumper)

    def volumes(self) -> Iterator[Volume]:
        """Returns an iterator over all volumes associated with this venue."""
        for anthology_id in self.item_ids:
            volume = self.parent.get_volume(anthology_id)
            if volume is None:
                raise ValueError(
                    f"Venue {self.id} lists associated volume {build_id_from_tuple(anthology_id)}, which doesn't exist"
                )
            yield volume


@define
class VenueIndex(SlottedDict[Venue]):
    """Index object through which venues and their associated volumes can be accessed.

    Provides dictionary-like functionality mapping venue IDs to [Venue][acl_anthology.venues.Venue] objects.

    Attributes:
        parent: The parent Anthology instance to which this index belongs.
        no_item_ids: If set to True, skips parsing all XML files, which means the reverse-indexing of Volumes via `Venue.item_ids` will not be available.
        is_data_loaded: A flag indicating whether the venue YAML files have been loaded and the index has been built.
    """

    parent: Anthology = field(repr=False, eq=False)
    no_item_ids: bool = field(repr=False, default=False)
    is_data_loaded: bool = field(init=False, repr=True, default=False)

    def load(self) -> None:
        """Loads and parses the `venues/*.yaml` files.

        Raises:
            KeyError: If a mandatory key is missing in a YAML file.
        """
        # This function exists so we can later add the option to read the index
        # from a cache if it doesn't need re-building.
        if self.is_data_loaded:
            return
        for yamlpath in self.parent.datadir.glob("yaml/venues/*.yaml"):
            venue = Venue.load_from_yaml(yamlpath, self.parent)
            self.data[venue.id] = venue
        self.build()
        self.is_data_loaded = True

    def reset(self) -> None:
        """Resets the index."""
        self.data = {}
        self.is_data_loaded = False

    def build(self) -> None:
        """Load the entire Anthology data and build an index of venues.

        Raises:
            ValueError: If a volume lists a venue ID that doesn't exist (i.e., isn't defined in the venue YAML files).
        """
        if self.no_item_ids:
            return
        for volume in self.parent.volumes():
            for venue_id in volume.venue_ids:
                if venue_id not in self.data:
                    raise ValueError(
                        f"Volume {volume.full_id} lists associated venue {venue_id}, which doesn't exist"
                    )
                self.data[venue_id].item_ids.append(volume.full_id_tuple)

    def save(self) -> None:
        """Saves all venue metadata to `venues/*.yaml` files."""
        for venue in self.values():
            venue.save()



# Copyright 2023-2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Constants used within the library."""

from datetime import date
from typing import Final

FRONTMATTER_ID: Final = "0"
"""Paper ID that signals that a paper is frontmatter."""

NO_BIBKEY: Final = "__NO_BIBKEY__"
"""Sentinel value to be set on a newly created paper, before its bibkeys is generated."""

UNKNOWN_INGEST_DATE = date(1900, 1, 1)
"""Default ingestion date."""

RE_EVENT_ID = r"^[a-z0-9]+-[0-9]{4}$"
"""A regular expression matching a valid event ID."""

RE_ISO_DATE = r"^[0-9]{4}-[0-9]{2}-[0-9]{2}$"
"""A regular expression matching a date in ISO format."""



# Copyright 2023-2025 Marcel Bollmann <marcel@bollmann.me>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import sys
from attrs import define, field, validators as v
from lxml import etree
from pathlib import Path
from typing import Any, Iterator, Optional, cast, TYPE_CHECKING

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from ..containers import SlottedDict
from ..exceptions import AnthologyDuplicateIDError, AnthologyInvalidIDError
from ..text.markuptext import MarkupText
from ..utils.attrs import auto_validate_types, int_to_str
from ..utils.ids import infer_year, is_valid_collection_id
from ..utils.logging import get_logger
from ..utils import xml
from .event import Event
from .types import EventLinkingType, VolumeType
from .volume import Volume
from .paper import Paper

if TYPE_CHECKING:
    from _typeshed import StrPath
    from ..anthology import Anthology
    from .index import CollectionIndex


log = get_logger()


@define(field_transformer=auto_validate_types)
class Collection(SlottedDict[Volume]):
    """A collection of volumes and events, corresponding to an XML file in the `data/xml/` directory of the Anthology repo.

    Provides dictionary-like functionality mapping volume IDs to [Volume][acl_anthology.collections.volume.Volume] objects in the collection.

    Info:
        To create a new collection, use [`CollectionIndex.create()`][acl_anthology.collections.index.CollectionIndex.create].

    Attributes: Required Attributes:
        id: The ID of this collection (e.g. "L06" or "2022.emnlp").
        parent: The parent CollectionIndex instance to which this collection belongs.
        path: The path of the XML file representing this collection.

    Attributes: Non-Init Attributes:
        event: An event represented by this collection.
        is_data_loaded: A flag indicating whether the XML file has already been loaded.
    """

    id: str = field(converter=int_to_str)
    parent: CollectionIndex = field(repr=False, eq=False)
    path: Path = field(converter=Path)
    event: Optional[Event] = field(
        init=False,
        repr=False,
        default=None,
        validator=v.optional(v.instance_of(Event)),
    )
    is_data_loaded: bool = field(init=False, repr=True, default=False)

    @id.validator
    def _check_id(self, _: Any, value: str) -> None:
        if not is_valid_collection_id(value):
            raise AnthologyInvalidIDError(value, "Not a valid Collection ID")

    @property
    def root(self) -> Anthology:
        """The Anthology instance to which this object belongs."""
        return self.parent.parent

    def volumes(self) -> Iterator[Volume]:
        """An iterator over all Volume objects in this collection."""
        if not self.is_data_loaded:
            self.load()
        yield from self.data.values()

    def papers(self) -> Iterator[Paper]:
        """An iterator over all Paper objects in all volumes in this collection."""
        for volume in self.volumes():
            yield from volume.papers()

    def get_event(self) -> Optional[Event]:
        """An Event explicitly defined in this collection, if any."""
        if not self.is_data_loaded:
            self.load()
        return self.event

    def _add_volume_from_xml(self, meta: etree._Element) -> Volume:
        """Creates a new volume belonging to this collection.

        Parameters:
            meta: The `<meta>` element for the volume.

        Returns:
            The created volume.

        Raises:
            AnthologyDuplicateIDError: If a volume with the given ID already exists.
        """
        volume = Volume.from_xml(self, meta)
        if volume.id in self.data:
            raise AnthologyDuplicateIDError(
                volume.id, "Volume already exists in collection {self.id}"
            )
        self.data[volume.id] = volume
        return volume

    def _set_event_from_xml(self, meta: etree._Element) -> None:
        """Creates and sets a new event belonging to this collection.

        Parameters:
            meta: The `<event>` element.

        Raises:
            ValueError: If an event had already been set; collections may only have a single event.
        """
        if self.event is not None:
            raise ValueError(f"Event already defined in collection {self.id}")
        self.event = Event.from_xml(self, meta)

    def validate_schema(self) -> Self:
        """Validates the XML file belonging to this collection against the RelaxNG schema.

        Raises:
            lxml.etree.DocumentInvalid: If the XML file does not validate against the schema.
        """
        self.root.relaxng.assertValid(etree.parse(self.path))
        return self

    def create_volume(
        self,
        id: str,
        title: MarkupText,
        year: Optional[str] = None,
        type: VolumeType = VolumeType.PROCEEDINGS,
        **kwargs: Any,
    ) -> Volume:
        """Create a new [Volume][acl_anthology.collections.volume.Volume] object in this collection.

        Parameters:
            id: The ID of the new volume.
            title: The title of the new volume.
            year: The year of the new volume (optional); if None, will infer the year from this collection's ID.
            type: Whether this is a journal or proceedings volume; defaults to [VolumeType.PROCEEDINGS][acl_anthology.collections.types.VolumeType].
            **kwargs: Any valid list or optional attribute of [Volume][acl_anthology.collections.volume.Volume].

        Returns:
            The created [Volume][acl_anthology.collections.volume.Volume] object.

        Raises:
            AnthologyDuplicateIDError: If a volume with the given ID already exists.
            ValueError: If this collection has an old-style ID.
        """
        if not self.is_data_loaded:
            self.load()
        if not self.id[0].isdigit():
            raise ValueError(
                f"Can't create volume in collection {self.id} with old-style ID"
            )
        if id in self.data:
            raise AnthologyDuplicateIDError(
                id, f"Volume already exists in collection {self.id}"
            )

        if year is None:
            year = infer_year(self.id)

        volume = Volume(
            id=id,
            parent=self,
            booktitle=title,
            year=year,
            type=type,
            **kwargs,
        )
        volume.is_data_loaded = True

        # For convenience, if editors were given, we add them to the index here
        if volume.editors:
            self.root.people._add_to_index(volume.editors, volume.full_id_tuple)

        self.data[id] = volume
        return volume

    def create_event(
        self,
        id: Optional[str] = None,
        **kwargs: Any,
    ) -> Event:
        """Create a new (explicit) [Event][acl_anthology.collections.event.Event] object in this collection.

        Parameters:
            id: The ID of the event; must follow [`RE_EVENT_ID`][acl_anthology.constants.RE_EVENT_ID].  If None (default), and this collection has a new-style ID, will generate an event ID based on this (e.g., collection "2022.emnlp" will generate event "emnlp-2022").
            **kwargs: Any valid list or optional attribute of [Event][acl_anthology.collections.event.Event].

        Returns:
            The created [Event][acl_anthology.collections.event.Event] object.

        Raises:
            ValueError: If an explicitly defined event already exists in this collection, or if `id` was None and this collection has an old-style ID.

        Danger:
            If the [event index][acl_anthology.collections.eventindex.EventIndex] is loaded _and_ an event with the given ID is already implicitly defined, the newly created event will replace that one, _but will inherit its co-located IDs_.  It is currently not possible to explicitly create an event without also explicitly linking all co-located item IDs to it, but for performance reasons (this linking needs to load the entire Anthology data), it _only happens when the event index is loaded._  This means that e.g. entirely new proceedings can be created without the performance impact of loading everything, but for adding new events to existing proceedings, the event index should probably be loaded first.
        """
        if not self.is_data_loaded:
            self.load()
        if self.event is not None:
            raise ValueError(
                f"Can't create event in collection {self.id}: already exists"
            )
        if id is None:
            if not self.id[0].isdigit():
                raise ValueError(
                    f"Can't create event in collection {self.id} without an explicitly given ID"
                )
            id = "-".join(self.id.split(".")[::-1])

        self.event = Event(
            id=id,
            parent=self,
            is_explicit=True,
            **kwargs,
        )
        self.root.events._add_to_index(self.event)
        return self.event

    def load(self) -> None:
        """Loads the XML file belonging to this collection."""
        if self.is_data_loaded:
            return

        log.debug(f"Parsing XML data file: {self.path}")
        current_volume = cast(Volume, None)  # noqa: F841
        for _, element in etree.iterparse(
            self.path,
            tag=("meta", "frontmatter", "paper", "volume", "event", "collection"),
        ):
            discard_element = True
            if (
                element.tag == "meta"
                and (parent := element.getparent()) is not None
                and parent.tag != "event"
            ):
                # Seeing a volume's <meta> block instantiates a new volume
                current_volume = self._add_volume_from_xml(element)  # noqa: F841
            elif element.tag in ("paper", "frontmatter"):
                current_volume._add_paper_from_xml(element)
            elif element.tag == "volume":
                current_volume = cast(Volume, None)  # noqa: F481
            elif element.tag == "event":
                self._set_event_from_xml(element)
                element.clear()
            elif element.tag == "collection":
                if element.get("id") != self.id:
                    raise AnthologyInvalidIDError(
                        element.get("id"),
                        f"File {self.path} does not contain Collection {self.id}",
                    )
            else:
                # Keep element around; should only apply to <event><meta> ...
                discard_element = False

            if discard_element:
                element.clear()

        if self.event is not None:
            # Events are implicitly linked to volumes defined in the same collection
            self.event.colocated_ids = [
                (volume.full_id_tuple, EventLinkingType.INFERRED)
                for volume in self.data.values()
                # Edge case: in case the <colocated> block lists a volume in
                # the same collection, don't add it twice
                if (volume.full_id_tuple, EventLinkingType.EXPLICIT)
                not in self.event.colocated_ids
            ] + self.event.colocated_ids

        self.is_data_loaded = True

    def save(self, path: Optional[StrPath] = None, minimal_diff: bool = True) -> None:
        """Saves this collection as an XML file.

        Arguments:
            path: The filename to save to. If None, defaults to `self.path`.
            minimal_diff: If True (default), will compare against an existing XML file in `self.path` to minimize the difference, i.e., to prevent noise from changes in the XML that make no semantic difference.  See [`utils.xml.ensure_minimal_diff`][acl_anthology.utils.xml.ensure_minimal_diff] for details.
        """
        if path is None:
            path = self.path
        collection = etree.Element("collection", {"id": self.id})
        for volume in self.volumes():
            collection.append(volume.to_xml(with_papers=True))
        if self.event is not None and self.event.is_explicit:
            collection.append(self.event.to_xml())
        if self.path.is_file() and minimal_diff:
            reference = etree.parse(self.path).getroot()
            xml.indent(collection)  # allows for better checking for equivalence
            xml.ensure_minimal_diff(collection, reference)
        xml.indent(collection)
        with open(path, "wb") as f:
            f.write(etree.tostring(collection, xml_declaration=True, encoding="UTF-8"))
